//
// Copyright 1998 Omni Development, Inc.  All rights reserved.
//
// Author: Timothy J. Wood <tjw@omnigroup.com>
// Date:   Apr 19, 1998
//
// You may do what ever you want with this example code.  No warranty expressed or implied.
//

#import "OFPageMonitor.h"

#import <mach/mach.h>
#import <mach/exception.h>
#import <mach/cthreads.h>
#import <mach/mach_error.h>
#import <mach/exc_server.h>

typedef struct _OFPageMonitorThreadState {
    NSLock         *lock;
    port_t          exc_port;
    port_t          clear_port;
    port_t          old_exc_port;
    vm_address_t    baseAddress;
    vm_size_t       size;
    cthread_t       thread;
    void          **modifiedPages;
    unsigned int    modifiedPageCount;
    unsigned int    modifiedPageMax;
} OFPageMonitorThreadState;

static any_t exc_thread(OFPageMonitorThreadState *threadState);


@implementation OFPageMonitor

- initWithAddress: (void *) address
           length: (unsigned long) length;
{
    unsigned int  pageSize = NSPageSize();
    unsigned int  pageMask = (pageSize - 1);
    vm_address_t  regionAddress, regionEnd;
    vm_size_t     regionSize;
    vm_prot_t     regionProtection, regionMaxProtection;
    
    kern_return_t rc;

    regionAddress = (vm_address_t)address;
    regionEnd     = (vm_address_t)address + length;

    // Verfiy alignment
    if (regionAddress & pageMask)
        [NSException raise: NSInvalidArgumentException
                    format: @"OFPageMonitor start address must be page aligned (0x%08x was given).", address];
    if (length & pageMask)
        [NSException raise: NSInvalidArgumentException
                    format: @"OFPageMonitor length must be a multiple of NSPageSize() (0x%08x was given).", length];

    // Make sure that the input address range really is writable
    while (regionAddress < regionEnd) {
        vm_inherit_t  regionInheritance;
        boolean_t     regionIsShared;
        port_t        regionPort;
        vm_offset_t   regionOffset;

        rc = vm_region(task_self(), &regionAddress, &regionSize,
                       &regionProtection, &regionMaxProtection,
                       &regionInheritance, &regionIsShared, &regionPort, &regionOffset);
        if (rc != KERN_SUCCESS) {
            mach_error("vm_region", rc);
            exit(1);
        }

        // For right now, we'll assume that this will only be called on regions that are read/write/no-exec.
        // If this wasn't the case (like if you had a rwx or a write-only region for some weird reason),
        // then you'd need to maintain a list of what the original permisions were.  This shouldn't
        // happen in practice, but it is worth noting.

        if (regionProtection != (VM_PROT_READ | VM_PROT_WRITE))
            [NSException raise: NSInvalidArgumentException
                        format: @"OFPageMonitory must be used on a region of memory that is read/write/no-exec protections (protection was %d).", regionProtection];

        regionAddress += regionSize;
    }

    // Start a thread to listen for Mach exceptions occuring due to invalid writes.  Note that each
    // OFPageMonitory consumes a Mach thread.  This is kind of expensive -- a more advanced solution
    // would use a single thread to monitor all of the address regions for all OFPageMonitors.
    // This would require some cool interlocks to modify the list of address ranges.
    // Another thing to note is that right now OFPageMonitor will not work reliably if two monitors
    // are watching the same region.  In fact, it isn't even clear if two page monitors can exist
    // on two different regions (due the the chaining of exception handler ports).  It might work.
    // I haven't tested it.

    threadState = NSZoneMalloc(NULL, sizeof(*threadState));
    threadState->lock = [[NSLock alloc] init];
    threadState->baseAddress = (vm_address_t)address;
    threadState->size        = length;
    threadState->modifiedPageCount = 0;
    threadState->modifiedPageMax   = 16;
    threadState->modifiedPages     = NSZoneMalloc(NULL, sizeof(*threadState->modifiedPages) * threadState->modifiedPageMax);
    
    // Get the old exception port for this thread (our reply port) and the task exception port
    threadState->clear_port = thread_reply();
    rc = task_get_exception_port(task_self(), &threadState->old_exc_port);
    if (rc != KERN_SUCCESS) {
        mach_error("task_get_exception_port", rc);
        exit(1);
    }

    // Create a new exception port for this this thread
    rc = port_allocate(task_self(), &threadState->exc_port);
    if (rc != KERN_SUCCESS) {
        mach_error("port_allocate", rc);
        exit(1);
    }
    
    // Set our exception port -- any exception that happen from now until the catch thread is ready will cause this thread to get suspended.
    rc = thread_set_exception_port(thread_self(), threadState->exc_port);
    if (rc != KERN_SUCCESS) {
        mach_error("thread_set_exception_port", rc);
        exit(1);
    }

    // Start the thread
    threadState->thread = cthread_fork((cthread_fn_t)exc_thread, (any_t)threadState);

    // Now, make the region unwritable.  If an error happens here then we have a thread running that we need to kill.
    // For now, we'll just bomb out.  We can't move this protection before the thread is started very cleanly.
    // We'd have to put it before the exception port for this thread is set.  At a minimum, we'd need to set
    // that port back to the original.  At best, we'd need to temporarily install a thread that would move messages
    // from the old port to the new port.  That's probably overkill, though.
    rc = vm_protect(task_self(), threadState->baseAddress, threadState->size, NO, VM_PROT_READ);
    if (rc != KERN_SUCCESS) {
        mach_error("vm_protect", rc);
        exit(1);
    }
    
    return self;
}


- (void) dealloc;
{
    kern_return_t rc;
    
    // Restore permissions on all pages
    rc = vm_protect(task_self(), threadState->baseAddress, threadState->size, NO, VM_PROT_READ | VM_PROT_WRITE);
    if (rc != KERN_SUCCESS) {
        mach_error("vm_protect", rc);
        exit(1);
    }

    // Reset the exception port back to the old port.  Any new exceptions should
    // go to the default (or old) handler now.
    rc = thread_set_exception_port(thread_self(), threadState->old_exc_port);
    if (rc != KERN_SUCCESS) {
        mach_error("thread_set_exception_port", rc);
        exit(1);
    }

    // Destroy the watch thread.  If were were in the middle of handling an exception, this
    // could do bad stuff.
    cthread_exit(threadState->thread);

    // Toss the old exception port
    rc = port_deallocate(task_self(), threadState->exc_port);
    if (rc != KERN_SUCCESS) {
        mach_error("port_deallocate", rc);
        exit(1);
    }

    [threadState->lock release];
    NSZoneFree(NULL, threadState->modifiedPages);
    NSZoneFree(NULL, threadState);

    [super dealloc];
}

- (void **) modifiedPages;
{
    void         **pages;
    unsigned int  size;
    
    // Make sure that additional incoming writes don't fnorkle out getting the list of modified pages
    [threadState->lock lock];

    size = sizeof(void *) * threadState->modifiedPageCount;
    pages = NSZoneMalloc(NULL, size + sizeof(void *));
    memmove(pages, threadState->modifiedPages, size);
    pages[threadState->modifiedPageCount] = NULL;
        
    [threadState->lock unlock];

    return pages;
}

@end






static OFPageMonitorThreadState *currentThreadState = NULL;

/* Listen on the exception port. */
any_t exc_thread(OFPageMonitorThreadState *threadState)
{
    kern_return_t   r;
    char           *msg_data[2][64];
    msg_header_t   *imsg = (msg_header_t *)msg_data[0],
                   *omsg = (msg_header_t *)msg_data[1];

    if (currentThreadState) {
        printf("OFPageMonitor doesn't currently support multiple instances\n");
        exit(1);
    }
    // We need to communicate the threadState to catch_exception_raise.
    // We could put a lock around the call to exc_server, but I'm lazy right now.
    // Also, it might cause some weird deadlocks or something -- I dunno.
    currentThreadState = threadState;
    
    /* Wait for exceptions. */
    while (1) {
        imsg->msg_size = 64;
        imsg->msg_local_port = threadState->exc_port;
        r = msg_receive(imsg, MSG_OPTION_NONE, 0);

        if (r==RCV_SUCCESS) {
            //fprintf(stderr, "Got a message\n");

            /* Give the message to the Mach exception server. */
            if (exc_server(imsg, omsg)) {
                /* Send the reply message that exc_serv gave us. */
                r = msg_send(omsg, MSG_OPTION_NONE, 0);
                if (r != SEND_SUCCESS) {
                    mach_error("exc_thread msg_send", r);
                    exit(1);
                }
            } else { /* exc_server refused to handle imsg. */
                // NSLog is reported to be thread safe in 4.2 and later
                NSLog(@"exc_server didn't like the message");
                exit(2);
            }
        } else { /* msg_receive() returned an error. */
            mach_error("exc_thread msg_receive", r);
            exit(3);
        }
    }
}

/*
 * catch_exception_raise() is called by exc_server(). If this routine
 * doesn't exist, we'll get a IOT trap and exit.
 */
kern_return_t catch_exception_raise(port_t exception_port,
    port_t thread, port_t task, int exception, int code, int subcode)
{
#ifdef DEBUG
    fprintf(stderr,
            "catch_exception_raise called (exception = 0x%x, code = 0x%x, subcode = 0x%x)\n",
            exception, code, subcode);
#endif
    
    if (exception == EXC_BAD_ACCESS) {
        // subcode is the address that we failed on.  We'll have to fix that address in order for
        // the other thread to continue w/o immediately getting another exception

#ifdef DEBUG
        fprintf(stderr, "exception = EXC_BAD_ACCESS\n");
#endif
        // First, is this an address that we are supposed to be watching?
        if (currentThreadState->baseAddress <= subcode && subcode < currentThreadState->baseAddress + currentThreadState->size) {
            vm_address_t  page;
            kern_return_t rc;

#ifdef DEBUG
            fprintf(stderr, "\taddress in range\n");
#endif
            // Compute the page for this address
            page = subcode & ~(vm_page_size - 1);
                
            // Record it -- shouldn't be duplicates since we immediately turn on write access (although if
            // you have one page monitor working when multiple threads are writing into a block of memory
            // there might be.  We'll ignore it and let the duplicates happen.
            [currentThreadState->lock lock];
            
            if (currentThreadState->modifiedPageCount == currentThreadState->modifiedPageMax) {
                currentThreadState->modifiedPageMax *= 2;
                currentThreadState->modifiedPages = NSZoneRealloc(NULL, currentThreadState->modifiedPages,
                                                                  currentThreadState->modifiedPageMax * sizeof(*currentThreadState->modifiedPages));
            }
            currentThreadState->modifiedPages[currentThreadState->modifiedPageCount] = (void *)page;
            currentThreadState->modifiedPageCount++;

            // Make the page writeable again.  This prevents other writes the page from going through this
            // code (unless we are dealing with a multi-thread app).  At a minimum, this should keep the number
            // of hits way down.
            rc = vm_protect(task_self(), page, vm_page_size, NO, VM_PROT_READ | VM_PROT_WRITE);
            if (rc != KERN_SUCCESS) {
                mach_error("vm_protect", rc);
                exit(1);
            }

            [currentThreadState->lock unlock];
        }

        return KERN_SUCCESS;
    }
    
    return KERN_FAILURE;
}
